/************************************************************************************/
/* Copyright ?????????                                                              */
/* All rights reserved.                                                             */  
/*                                                                                  */
/************************************************************************************/
/*                                                                                  */
/* AUTHOR: Aby Louw                                                                 */
/* Email:                                                                           */
/* MODIFIED: Feb 10, 2004                                                           */
/*                                                                                  */
/************************************************************************************/

#include <stdio.h>
#include "EST.h"
#include "EST_types.h"
#include "EST_error.h"


static float BIG_NUMBER = 9999;

typedef enum {NoTone=0,M,S,T,H,U,B,L,D} tones;


float octave(float t)
{
    return log(t)/log(2);
}

float linear(float t)
{
    return pow(2,t);
}

int round(float t)
{
    return int(t + 0.5);
}


float make_estimate(int tone_current, float *last_target, float mid, float top,
		    float bottom, float HIGHER, float LOWER, float UP, float DOWN)
{
    tones current_tone = tones(tone_current);
    double est = 0;

    switch(current_tone) 
    {
      case M:
	est = mid;
	break;

      case S:
	est = (*last_target);
	break;

      case T:
	est = top;
	break;
	
      case H:
	est = (*last_target) + (top - (*last_target)) * HIGHER;
	break;
    
      case U:
	est = (*last_target) + (top - (*last_target)) * UP;
	break;
    
      case B:
	est = bottom;
	break; 
    
      case L:
	est = (*last_target) - ((*last_target) - bottom) * LOWER;
	break; 
	
      case D:
	est = (*last_target) - ((*last_target) - bottom) * DOWN;
	break; 
    }
    return est;
}



void optimise(EST_DVector f, EST_DVector t, float mid, float range, float *best_mid, float *best_range,
	      EST_IVector &best_intsint, EST_DVector &best_estimate, int nval,float *min_error2, float HIGHER,
	      float LOWER, float UP, float DOWN, float MIN_PAUSE)
{
    float top = mid + range/2.0;
    float bottom = mid - range/2.0;
    float f0 = f.a_check(0);

    EST_IVector intsint;
    intsint.resize(nval);

    if ((top - f0) < fabs(f0 - mid))
    {
	intsint.a_check(0) = tones(T);
    }
    else if ((f0 - bottom) < fabs(f0 - mid))
    {
	intsint.a_check(0) = tones(B);
    }
    else 
    { 
	intsint.a_check(0) = tones(M);
    }
    
    float last_estimate = 0;

    float testimate = make_estimate(intsint.a_check(0),&last_estimate,mid,top,bottom,HIGHER,LOWER,UP,DOWN);
    
    EST_DVector estimate;
    estimate.resize(nval);
    
    estimate.a_check(0) = testimate;
    float error = fabs(testimate - f0);
    float ss_error = error*error;
    last_estimate = testimate;
    
    for (int i = 1; i < nval; i++)
    {
	float target = f.a_check(i);

	// after pause choose from [MTB]
	if ((t.a_check(i) - t.a_check(i-1)) > MIN_PAUSE) 
	{
	    if ((top - target) < fabs(target - mid))
	    {
		intsint.a_check(i) = tones(T);
	    }
	    else if ((target - bottom) < fabs(target - mid))
	    {
		intsint.a_check(i) = tones(B);
	    }
	    else
	    {
		intsint.a_check(i) = tones(M);
	    }
	    // #elsewhere any tone except M
 	} 
	else 
	{ 
 	    float min_difference = BIG_NUMBER;
	    int best_tone = tones(NoTone);
	    
	    for (int j = 1; j < 9; j++)
	    {
		if (j != tones(M))
		{
		    testimate = make_estimate(j,&last_estimate,mid,top,bottom,HIGHER,LOWER,UP,DOWN);
		    float difference = fabs(target - testimate);
		    
		    if (difference < min_difference) 
		    {
			min_difference = difference;
			best_tone = j;
		    } 
		} 
	    }
	    intsint.a_check(i) = best_tone;
 	} 
	
	estimate.a_check(i) = make_estimate(intsint.a_check(i),&last_estimate,mid,top,bottom,HIGHER,LOWER,UP,DOWN);
 	error = fabs(estimate.a_check(i) - f.a_check(i));
	ss_error += error*error;
 	last_estimate = estimate.a_check(i);
    } 
    
    if (ss_error < (*min_error2)) 
    {
 	(*min_error2) = ss_error;
 	(*best_range) = range;
	(*best_mid) = mid;
	best_intsint = intsint;
	best_estimate = estimate;
    }
}



int main(int argc, char *argv[]) 
{ 
    EST_StrList files;
    EST_Option cmd_line;

    /* parameters for optimisation */
    float MIN_PAUSE = 0.3;
    float MIN_RANGE = 0.5;
    float MAX_RANGE = 2.5;
    float STEP_RANGE = 0.1;
    float MEAN_SHIFT = 50;
    float STEP_SHIFT=1;
    
    /* parameters for target estimation (distance towards T or B respectively) */
    float HIGHER = 0.5;
    float LOWER = 0.5;
    float UP = 0.25;
    float DOWN = 0.25;
  
    if (argc == 1)
    {
	cout <<"use -h for Options help"<<endl;
	exit(0);
    }

    parse_command_line(
	argc, argv, 
	EST_String("-i [input f0 file] -o [output file] [options]\n")+
	"Summary: get momel marks from f0 track\n"
	" options :\n"
	"-h                   Options help.\n"
	"-i          <string> Input momel filename.\n"
	"-o          <string> Output filename for INTSINT labels.\n"
 	"-min_pause  <float>  Minimum pause length [s].                               Default 0.3\n"
	"-min_range  <float>  Minimum search range [Octave].                          Default 0.5\n"
	"-max_range  <float>  Maximum search range [Octave].                          Default 2.5\n"
	"-step_range <float>  Step of range search [Octave].                          Default 0.1\n"
	"-mean_shift <float>  Difference between max and mid f0, mid and min f0 [Hz]. Default 50\n"
	"-step_shift <float>  Step of f0 range search [Hz].                           Default 1\n"
	"-higher     <float>  INTSINT Higher label factor.                            Default 0.5\n"
	"-lower      <float>  INTSINT Lower label factor.                             Default 0.5\n"
	"-up         <float>  INTSINT Up label factor.                                Default 0.25\n"
	"-down       <float>  INTSINT Down label factor.                              Default 0.25\n",
	files, cmd_line);

 
    /****************************************/
    /* Check for input file                 */
    /****************************************/
    if (!cmd_line.present("-i"))
    {
	EST_error("no input file specified\n");
    }

    /****************************************/
    /* Check for output file                */
    /****************************************/
    if (!cmd_line.present("-o"))
    {
	EST_error("no output file specified\n");
    }

    /****************************************/
    /* Check for pause length               */
    /****************************************/
    if (cmd_line.present("-min_pause"))
    {
	MIN_PAUSE = cmd_line.fval("-min_pause");
    }

    /****************************************/
    /* Check for min range                  */
    /****************************************/
    if (cmd_line.present("-min_range"))
    {
	MIN_RANGE = cmd_line.fval("-min_range");
    }

    /****************************************/
    /* Check for max range                  */
    /****************************************/
    if (cmd_line.present("-max_range"))
    {
	MAX_RANGE = cmd_line.fval("-max_range");
    }

    /****************************************/
    /* Check for step range                  */
    /****************************************/
    if (cmd_line.present("-step_range"))
    {
	STEP_RANGE = cmd_line.fval("-step_range");
    }

    /****************************************/
    /* Check for mean shift                 */
    /****************************************/
    if (cmd_line.present("-mean_shift"))
    {
	MEAN_SHIFT = cmd_line.fval("-mean_shift");
    }
    
    /****************************************/
    /* Check for step shift                 */
    /****************************************/
    if (cmd_line.present("-step_shift"))
    {
	STEP_SHIFT = cmd_line.fval("-step_shift");
    }

    /****************************************/
    /* Check for HIGHER                     */
    /****************************************/
    if (cmd_line.present("-higher"))
    {
	HIGHER = cmd_line.fval("-higher");
    }

    /****************************************/
    /* Check for LOWER                      */
    /****************************************/
    if (cmd_line.present("-lower"))
    {
	LOWER = cmd_line.fval("-lower");
    }

    /****************************************/
    /* Check for UP                         */
    /****************************************/
    if (cmd_line.present("-up"))
    {
	UP = cmd_line.fval("-up");
    }
    
    /****************************************/
    /* Check for DOWN                       */
    /****************************************/
    if (cmd_line.present("-down"))
    {
	DOWN = cmd_line.fval("-down");
    }
    

    /****************************************/
    /* Load input into a matrix             */
    /****************************************/
    EST_Track momel;

    if (momel.load(cmd_line.sval("-i")) != read_ok)
	EST_error("could not load momel file \"%s\"\n",cmd_line.sval("-i").str());


    int nval = 0;
  

    /* load f0 and time vectors */
    EST_DVector f;
    EST_DVector t;

    nval = momel.num_frames();
  
    f.resize(nval);
    t.resize(nval);

    double sumf0 = 0;

    for (int i = 0; i < nval; i++)
    {
	f.a_check(i) = octave(momel.a(i,0));
	t.a_check(i) = momel.t(i);
	sumf0 += f.a_check(i);
	
 // 	printf("t = %f f0 = %f\n",t.a_check(i),f0.a_check(i));
    }

    float meanf0 = 0;

    meanf0 = sumf0/nval;
 
    float linear_mean_f0 = 0;

    linear_mean_f0 = round(linear(meanf0));

    float min_mean = 0;
    
    min_mean = linear_mean_f0 - MEAN_SHIFT;
    
    float max_mean = 0;
    max_mean = linear_mean_f0 + MEAN_SHIFT;

    float min_error2 = 0;
    
    min_error2 = BIG_NUMBER;
    
 
    int range_to = (int)((MAX_RANGE - MIN_RANGE)/STEP_RANGE + 0.5);
    int lm_to = (int)((max_mean - min_mean)/STEP_SHIFT + 0.5);

    float best_mid = 0;
    float best_range = 0;
    float mid = 0;
    EST_DVector best_estimate;
    best_estimate.resize(nval);
    
    EST_IVector best_intsint;
    best_intsint.resize(nval);

    
    
    for (int i = 0; i < range_to; i++) 
    {
	for (int j = 0; j < lm_to; j++) 
	{
	    float range = (MIN_RANGE + STEP_RANGE*i);
	    float lm = (min_mean + STEP_SHIFT*j);
	    mid = octave(lm);
	    optimise(f,t,mid,range,&best_mid,&best_range,best_intsint,best_estimate, nval,&min_error2,HIGHER,LOWER,UP,DOWN,MIN_PAUSE);
	} 
    } 

   
    mid = best_mid;
    float key = linear(mid);
    float bottom = round(linear(mid - best_range/2.0));
    float top = round(linear(mid + best_range/2.0));

    EST_Relation olabel;
    olabel.clear();
    olabel.f.set("top",top);
    olabel.f.set("bottom",bottom);
    olabel.f.set("mid",key);
    olabel.f.set("timing_style", "end");

    for (int i = 0; i < nval; i++)
    {
	EST_Item *item = olabel.append();
	item->set("end",t.a_check(i));

	switch(best_intsint.a_check(i)) 
	{
	  case 1:
	    item->set("name","M");
	    break;
	  case 2:
	    item->set("name","S");
	    break;
	  case 3:
	    item->set("name","T");
	    break;
	  case 4:
	    item->set("name","H");
	    break;
	  case 5:
	    item->set("name","U");
	    break;
	  case 6:
	    item->set("name","B");
	    break; 
          case 7:
	    item->set("name","L");
	    break; 
	  case 8:
	    item->set("name","D");
	    break; 
	}
    }

    if (olabel.save(cmd_line.sval("-o")) != write_ok)
    {
	EST_error("INTSINT: can't write olabel file \"%s\"\n",cmd_line.sval("-o").str());
    }
    
    return 0;
}
