/************************************************************************************/
/* Copyright ?????????                                                              */
/* All rights reserved.                                                             */  
/*                                                                                  */
/************************************************************************************/
/*                                                                                  */
/* AUTHOR: Aby Louw                                                                 */
/* Email:                                                                           */
/* MODIFIED: Feb 10, 2004                                                           */
/*                                                                                  */
/************************************************************************************/

#include <stdio.h>
#include "vectorwindow.h"
#include "momel.h"
#include "EST_cmd_line_options.h"




int main(int argc, char *argv[]) 
{ 
    EST_StrList files;
    EST_Option cmd_line;
          
    /* options variables */
    float delta = 5;                    /* Percentage of accepted error */
    int A = 30;                         /* Window size for generating targets [ms] */
    int R = 20;                         /* Window size for discriminating targets [ms] */
    int hzmin = 80;                     /* Minimum accepted f0 frequency [Hz] */
    int hzmax = 500;                    /* Maximum accepted f0 frequency [Hz] */
    int version = 1;                    /* momel [1] / new [2] */
    bool do_debug = false;              /* Display information on the values */
  
    if (argc == 1)
    {
	cout <<"use -h for Options help"<<endl;
	exit(0);
    }

    parse_command_line(
	argc, argv, 
	EST_String("-i [input f0 file] -o [output file] [options]\n")+
	"Summary: get momel marks from f0 track\n"
	" options :\n"
	"-h                     Options help.\n"
	"-i            <string> Input f0 filename.\n"
	"-itype        <string> Input file type (optional).\n"
	"-o            <string> Output filename for momel marks.\n"
	"-otype        <string> Output file type for momel marks.                              Default est\n"
	"-delta        <float>  Percentage of accepted error between MOMEL and initial curves. Default 5\n"
	"-A            <int>    Window size for generating targets [ms].                       Default 30\n"
	"-R            <int>    Window size for discriminating targets [ms].                   Default 20\n"
	"-hzmin        <int>    Minimum accepted f0 frequency [Hz].                            Default 80\n"
	"-hzmax        <int>    Maximum accepted f0 frequency [Hz].                            Default 500\n"
	"-version      <string> momel / new.                                                   Default momel\n"
	"-debug        <int>    Display debug information. 0 [no] / 1 [yes]                    Default 0\n"
	"-spline_file  <string> Filename for quadratic spline from momel points.               Default nofile\n"
	"-s_file_type  <string> File output type for quadratic spline file.                    Default est\n"+
	options_track_filetypes_long(),
	files, cmd_line);


    /*******************************/
    /* Get options from cmdline    */
    /*******************************/
    
     /* delta */
    if (cmd_line.present("-delta"))
   	delta = cmd_line.fval("-delta",0);

    /* A */
    if (cmd_line.present("-A"))
	A = cmd_line.ival("-A");

    /* R */
    if (cmd_line.present("-R"))
	R = cmd_line.ival("-R");

    /* hzmin */
    if (cmd_line.present("-hzmin"))
	hzmin = cmd_line.ival("-hzmin");

    /* hzmax */
    if (cmd_line.present("-hzmax"))
	hzmax = cmd_line.ival("-hzmax");

    /* version */
    if (cmd_line.present("-version"))
	if (cmd_line.sval("-version") == "momel")
	{
	    version = 1;
	}
	else if (cmd_line.sval("-version") == "new")
	{
	    version = 2;
	}
	else
	    EST_error("Unknown version"); 

    /* debug */
    if (cmd_line.present("-debug"))
    {
	if (cmd_line.ival("-debug") == 0)
	    do_debug = false;
	else
	    do_debug = true;
    }

    /* output momel points file */
    if (!cmd_line.present("-o"))
    {
	EST_error("no ouput file specified\n");
    }
    
   
    /****************************************/
    /* Check for input file                 */
    /****************************************/
    if (!cmd_line.present("-i"))
    {
	EST_error("no input file specified\n");
    }
   
    /***************************************************************/
    /* Initialize the momel object with the input variables        */
    /***************************************************************/
    Momel MyMomel(do_debug,delta,A,R,hzmin,hzmax,version);

    /* ouput momel params */
    if (do_debug)
	MyMomel.PrintVars();

    /***************************************************************/
    /* Read the input f0                                           */
    /***************************************************************/

    if (cmd_line.present("-itype"))
    {
 	if (cmd_line.val("-itype") == "ascii")
	{
	    MyMomel.GetF0fromAsciiFile(cmd_line.sval("-i"));
	}
	else
	{
	    MyMomel.GetF0fromTrackFile(cmd_line.sval("-i"));
	}
    }
    else
    {
	MyMomel.GetF0fromTrackFile(cmd_line.sval("-i"));
    }

    /***************************************************************/
    /*     Momel Algorithm: Pre-processing                         */
    /***************************************************************/
    MyMomel.pre_process();

    /***************************************************************/
    /*     Momel Algorithm: target candidate estimation            */
    /***************************************************************/	
    MyMomel.get_target_candidates();

    /***************************************************************/
    /*     Momel Algorithm: partitionning of target candidates     */
    /***************************************************************/
    MyMomel.partition();
    
    /***************************************************************/
    /*     Momel Algorithm: reduction of target candidates         */
    /***************************************************************/
    MyMomel.reduce();

    /***************************************************************/
    /*     Momel Algorithm: calculate the momel curve              */
    /***************************************************************/
    MyMomel.calc_momel_curve();

    /***************************************************************/
    /*     Save Momel points                                       */
    /***************************************************************/
     
    if (cmd_line.present("-otype"))
    {
	MyMomel.SaveMomelMarkers(cmd_line.sval("-o"),cmd_line.sval("-otype"));
    }
    else
    {
	MyMomel.SaveMomelMarkers(cmd_line.sval("-o"),"est");
    }

    /***************************************************************/
    /*     Save Momel spline curve for view in praat               */
    /***************************************************************/
    if (cmd_line.present("-spline_file"))
    {
	if (cmd_line.present("-s_file_type"))
	    MyMomel.SaveMomelSpline(cmd_line.sval("-spline_file"),cmd_line.sval("-s_file_type"));
	else
	    MyMomel.SaveMomelSpline(cmd_line.sval("-spline_file"),"est");
    }


    return 0;
}
