/************************************************************************************/
/* Copyright ?????????                                                              */
/* All rights reserved.                                                             */  
/*                                                                                  */
/************************************************************************************/
/*                                                                                  */
/* AUTHOR: Aby Louw                                                                 */
/* Email:                                                                           */
/* MODIFIED: Feb 10, 2004                                                           */
/*                                                                                  */
/************************************************************************************/

#include <stdio.h>
#include "vectorwindow.h"

VectorWindow::VectorWindow(EST_FVector orig, int center, int lth)
{
    lngth = getoddlength(lth);
    startx = center - (int)ceil((lngth-1)/2.0);
    endx = center + (int)ceil((lngth-1)/2.0);
    centerx = center;
      
    window.resize(lngth);
    window.fill(0);

    int i = 0; /* counter */

    for (i = 0; (i < lngth) && ((startx + i) < 0); i++)
	window.a_check(i) = 0;

    for (; (i < lngth) && ((startx + i) < orig.length()); i++)
	window.a_check(i) = orig.a_check(startx + i);

    for(; i < lngth; i++)
	window.a_check(i) = 0;
}


/* copy constructor */
VectorWindow::VectorWindow(const VectorWindow& a):
    startx  (a.startx),
    endx    (a.endx),
    centerx (a.centerx),
    lngth   (a.lngth),
    window  (a.window)
{
}


// Desructor
VectorWindow::~VectorWindow()
{
}

int VectorWindow::getoddlength(int l)
{
    int a = (int)ceil(l/2.0);
    int b = (int)floor(l/2.0);

    int ret = 0;
    
    if (a==b)
    {
	/* number is even */
	ret = l + 1;
	return ret;
    }
    else
    {
	/* number is odd */
	ret = l;
	return ret;
    }
}


/* assign window to new (part of) vector */
void VectorWindow::assign(EST_FVector orig, int center, int lth)
{
    lngth = getoddlength(lth);
    startx = center - (int)ceil((lngth-1)/2.0);
    endx = center + (int)ceil((lngth-1)/2.0);
    centerx = center;
    
    window.resize(lngth);
    window.fill(0);
    
    for (int x = 0; x < lngth; x++)
    {
	if (((centerx + startx + x) > -1) && ((centerx + startx + x) < orig.length()))
	{
	    window.a_check(x) = orig.a_check(centerx + startx + x);
	}
    }  
}

/* get the value of the window at point x */
float VectorWindow::getvalue(int x)
{
    if ((x > endx) || (x < startx))
    {
	EST_warning("Tried to access column %d of %d column vector",
		    x-startx,lngth);
	return 0;
    }
    else
    {
	return window.a_check(x-startx);
    }
}

/* return window length */
int VectorWindow::length()
{
    return lngth;
}

/* return start value */
int VectorWindow::start()
{
    return startx;
}

/* return end value */
int VectorWindow::end()
{
    return endx;
}

/* return center value */
int VectorWindow::center()
{
    return centerx;
}

/* set the value of the window at point x */
void VectorWindow::setvalue(int x, float v)
{
    if ((x > endx) || (x < startx))
    {
	EST_warning("Tried to access column %d of %d column vector",
		    x-startx,lngth);
    }
    else
    {
	window.a_check(x-startx) = v;
    }
}


/* insert window into vector */
void VectorWindow::win2vector(EST_FVector& orig)
{
   for (int x = 0; x < lngth; x++)
   {
       if (((centerx + startx + x) > -1) && ((centerx + startx + x) < orig.length()))
       {
	   orig.a_check(centerx + startx + x) = window.a_check(x);
       }
   }
}
